"""Commands for manipulating local tweet collection."""

import sys

from poultry import consumers, options

dispatcher = options.Dispatcher()
command = dispatcher.command

# Making conda_build happy
dispatch = dispatcher.dispatch


@command()
def group(producer,
          file_name_template=('t', '%Y-%m-%d-%H.gz', ''),
          ):
    """Group tweets to files by date according to the template."""
    producer(consumers.to_tweet(consumers.group(file_name_template)))


@command()
def show(producer):
    """Print tweets in human readable form."""
    producer(consumers.to_tweet(consumers.show()))


@command()
def select(producer):
    """Print tweets as it is returned by the Twitter streaming API."""
    producer(consumers.print_())


@command()
def text(producer):
    """Print only tweet's text.

    It replaces the new line symbol (`\n`) with a space.

    """
    producer(consumers.to_tweet(consumers.print_text()))


@command()
def pprint(producer):
    """Pretty print tweet's json representation."""
    producer(consumers.to_tweet(consumers.pprint()))


@command()
def filter(producer, config):
    """Filter the tweets to files by filtering predicates defined in the configuration file."""
    dustbin_template = config.dustbin_template
    dustbin = consumers.group(dustbin_template) if dustbin_template is not None else None

    streams = tuple(
        (
            consumers.group(f.split_template),
            lambda c, _, f=f: c.filter(**f.predicates),
        )
        for f in config.filters
    )
    target = consumers.filter(streams, dustbin)

    producer(consumers.to_tweet(target))


@command()
def uniq(producer):
    """Omit repeated tweets."""
    producer(
        consumers.to_tweet(
            consumers.uniq(consumers.select())
        ),
    )


@command()
def timeline(producer, window=('w', '%Y-%m-%d-%H', '')):
    """Count the number of tweets per window."""

    producer(
        consumers.to_tweet(
            consumers.timeline(
                window=window,
                target=consumers.counter_printer(sys.stdout),
            ),
        ),
    )



import os
try:
    from ConfigParser import SafeConfigParser, NoOptionError, NoSectionError
except ImportError:
    from configparser import SafeConfigParser, NoOptionError, NoSectionError
from sys import version_info
PY2 = version_info < (3, )

from poultry.tweet import Coordinates, Tweet


class Config(object):

    default_config_file = os.path.expanduser('~/.poultry.cfg')

    def __init__(self, config_file):
        config = SafeConfigParser()
        config.read(config_file)
        self.config = config

    def sections(self):
        return self.config.sections()

    def items(self, sections):
        return self.config.items(sections)

    def get_elements(self, section, option):
        value = self.config.get(section, option)
        if PY2:
            value = value.decode('utf-8')
        return tuple(filter(None, (value.strip().split('\n'))))

    @property
    def dustbin_template(self):
        try:
            template = self.config.get('poultry', 'dustbin_template', raw=True)
        except NoSectionError:
            pass
        except NoOptionError:
            pass
        else:
            return template if template.strip() else None

    @property
    def _filter_sections(self):
        return (s for s in self.config.sections() if s.startswith('filter:'))

    @property
    def filters(self):
        #XXX would be nice to return a pair (s, self.filter(s))
        return (self.filter(s) for s in self._filter_sections)

    def filter(self, filter_name):
        # Gather follow and track predicates
        predicates = {p: self.get_elements(filter_name, p) for p in 'follow track'.split()}
        # Add locations
        predicates['locations'] = tuple(
            Coordinates.from_string(l) for l in self.get_elements(filter_name, 'locations')
        )
        split_template = self.config.get(filter_name, 'split_template', raw=True)
        try:
            start_date = self.config.get(filter_name, 'start_date')
        except NoOptionError:
            start_date = None

        try:
            end_date = self.config.get(filter_name, 'end_date')
        except NoOptionError:
            end_date = None

        return Filter(name=filter_name, split_template=split_template,
                      start_date=start_date, end_date=end_date,
                      **predicates)

    @property
    def global_filter(self):
        """The union of all the defined filters."""
        filters = iter(self.filters)
        result = next(filters)

        for filter_ in filters:
            result = result | filter_

        return result

    def merge(self, merge_name):
        inputs = self.get_elements(merge_name, 'inputs')
        return Merge(name=merge_name, inputs=inputs)

    @property
    def sentry_dsn(self):
        return self.config.get('fowler', 'sentry_dsn')


class Filter(object):
    """Twitter filter predicates.

    The predicates are::

      * https://dev.twitter.com/docs/streaming-api/methods#follow
      * https://dev.twitter.com/docs/streaming-api/methods#track
      * https://dev.twitter.com/docs/streaming-api/methods#locations

    """
    def __init__(self, follow=None, track=None, locations=None,
                 split_template=None, start_date=None, end_date=None,
                 name=None):
        self.follow = set(follow) if follow is not None else set()
        self.track = set(track) if track is not None else set()
        self.locations = set(locations) if locations is not None else set()

        self.split_template = split_template
        # XXX that's kind of a hack
        self.name = name.replace(':', '_') if name is not None else None

        self.start_date = Tweet._created_at_to_datetime(start_date) if start_date is not None else None
        self.end_date = Tweet._created_at_to_datetime(end_date) if end_date is not None else None

    def __or__(self, other):
        filtering_predicates = 'follow', 'track', 'locations'

        predicates = {p: self.predicates[p] for p in filtering_predicates}
        for p in filtering_predicates:
            predicates[p].update(other.predicates[p])

        return Filter(**predicates)

    @property
    def predicates(self):
        return {'follow': self.follow,
                'track': self.track,
                'locations': self.locations,
                'start_date': self.start_date,
                'end_date': self.end_date,
                }


class Merge(object):

    def __init__(self, name, inputs):
        self.name = name
        self.inputs = inputs




import fileinput
import contextlib

from poultry import consumers
from poultry.utils import get_file_names
from poultry.stream import from_twitter_api


def consume_stream(target, input_dir=None):
    """Read lines from the standard input or files in a directory.

    Behaves as a generator, should receive a .send() call to send a
    line to the target.

    """
    file_names = get_file_names(input_dir) if input_dir else []

    with contextlib.closing(fileinput.FileInput(file_names, openhook=fileinput.hook_compressed)) as lines:
        with consumers.closing(target):

            for line in lines:
                result = target.send(line)

                if result is not consumers.SendNext:
                    yield result


def from_stream(target, source=None, config=None):
    """Send lines from the standard input, the input directory or the Twitter Streaming API.

    :param target: a generator to which the read lines are sent.
    :param source: the path to a directory with tweet files.
    :param config: the config file

    """
    if source in ('twitter://sample', 'twitter://filter'):
        from_twitter_api(target, source, config)
    else:
        consumer = consume_stream(target, source)
        try:
            while True:
                next(consumer)
        except StopIteration:
            pass



from __future__ import print_function

import gzip
import logging
import sys
import time

from collections import OrderedDict, Counter
from contextlib import contextmanager
from itertools import chain
from pprint import pprint as _pprint

from poultry.tweet import Tweet, TweetValueError


logger = logging.getLogger(__name__)


def consumer(func):
    """A decorator function that takes care of starting a coroutine automatically on call.

    See http://www.dabeaz.com/generators/ for more details.

    """
    def start(*args, **kwargs):
        cr = func(*args, **kwargs)
        next(cr)
        return cr
    return start


@consumer
def nop(*args, **kwargs):
    item = None
    while True:
        item = yield item


@consumer
def show(template=u'{}\n'):
    """Print tweet text and meta information."""
    while True:
        print(template.format((yield)))


@consumer
def print_(output=None, template=u'{}\n'):
    """Print stripped items."""
    if output is None:
        output = sys.stdout
    while True:
        item = yield
        output.write(template.format((item).strip('\n')))


@consumer
def pprint():
    """Pretty print tweet's json object."""
    while True:
        tweet = (yield)
        _pprint(tweet.parsed)


@consumer
def print_text():
    """Print only tweet's text."""
    while True:
        tweet = (yield)
        print(tweet.text.replace('\n', ' '))


@consumer
def counter_printer(output=sys.stdout):
    while True:
        counter = yield

        for key, value in sorted(counter.items()):
            output.write('{} {}\n'.format(key, value))


@consumer
def to_tweet(target):
    """Convert the input items to tweets."""
    result = None

    with closing(target):
        while True:
            item = yield result

            try:
                tweet = Tweet(item)
            except TweetValueError:
                result = SendNext
            else:
                result = target.send(tweet)


@consumer
def group(file_name_template='%Y-%m-%d-%H.gz',
          max_open_files=1,
          ):
    """Group tweets to files by date according to the file_name_template."""
    files = OrderedDict()

    try:
        while True:
            tweet = yield

            created_at = tweet.created_at.strftime(file_name_template)

            try:
                f = files[created_at]
            except KeyError:
                print(created_at)

                if len(files) > max_open_files:
                    files = OrderedDict(sorted(files.items()))

                    _, first = files.popitem(last=False)
                    first.close()

                f = gzip.open(created_at, 'a')
                files[created_at] = f

            f.write("{t}\n".format(t=tweet.raw))

    finally:
        for f in files.values():
            f.close()


@consumer
def filter(streams, dustbin=None, send_to_all=True):
    """Filter items to flows by filtering predicates.

    :param streams: sequence of `(target, predicate)` pairs.

    :param send_to_all: if set to `False` then in case an item
                        satisfies several predicates it will be sent
                        to only one.

    `predicate` is a function: current, first --> Bool

    """
    def update_all_targets():
        all_targets = [t for t, _ in streams]
        if dustbin is not None:
            all_targets.append(dustbin)
        return all_targets

    all_targets = update_all_targets()

    with closing(mutable_targets=all_targets):
        current = first = yield

        while True:
            sent = False
            for target, predicate in streams:
                if predicate(current, first):
                    target.send(current)
                    sent = True

                    if not send_to_all:
                        break

            if dustbin is not None and not sent:
                dustbin.send(current)

            current = yield
            all_targets = update_all_targets()


@consumer
def uniq(target, seen_ids=None):
    """Omit repeated tweets."""
    seen_ids = set(seen_ids) if seen_ids is not None else set()

    with closing(target):
        while True:
            tweet = yield
            id_ = tweet.id

            if id_ not in seen_ids:
                seen_ids.add(id_)
                target.send(tweet)


@consumer
def split(*targets):
    """Send the input items to each target."""
    with closing(*targets):
        while True:
            item = yield

            for target in targets:
                target.send(item)


@consumer
def to_simple_queue(queue):
    """Put items to a simple queue."""
    while True:
        item = yield
        queue.put(item)


@consumer
def count(*counters, **kwargs):
    """Universal element counter.

    :param counters: counters to update.

    :param target: an optional target to which the cached local counts
                   are sent per each batch.

    :param provider: a function which provides elements that have to
                     be counted given an input item.

    """
    provider = kwargs.get('provider', None)
    target = kwargs.get('target', None)
    targets = [target] if target else []

    if provider is None:
        provider = lambda x: [x]

    if not counters:
        counters = Counter(),

    with closing(*targets):
        while True:
            with lazy_counter(*counters, target=target) as c:
                while True:
                    item = yield
                    c.update(provider(item))


def count_tokens(*counters, **kwargs):
    """Count tokens in a tweet."""
    if kwargs.pop('distinguish_hashtags', True):
        kwargs['provider'] = lambda tweet: chain(tweet.tokens,
                                                 (u'#{}'.format(h) for h in tweet.hashtags),
                                                 )
    else:
        kwargs['provider'] = lambda tweet: chain(tweet.tokens, tweet.hashtags)

    return count(*counters, **kwargs)


def timeline(*counters, **kwargs):
    """Count tweet's creation time."""
    window = kwargs.pop('window', '%Y-%m-%d-%H')
    kwargs['provider'] = lambda tweet: [tweet.created_at.strftime(window)]

    return count(*counters, **kwargs)


@consumer
def batch(target, flow_name=None, splitter=None):
    """Batch a stream of tweets to chunks defined by `splitter`.

    :param target: a coroutine tweets are sent to.
    :param flow_name: an optional flow name, which is used mainly for logging.
    :param splitter: a function which decides whether a new batch has started.
    :type splitter: (datetime, Tweet) --> bool

    """
    batch_size = 0

    if splitter is None:
        def splitter(current, first):
            return current != first

    with closing(target):

        current = first = yield

        while True:
            batch_size += 1

            if splitter(current, first):
                logger.debug('Sending a batch of %s items. (%s) '
                             'first: %s current: %s',
                             batch_size, flow_name, first, current)

                target.throw(BatchEndException(current, batch_size))
                batch_size = 0
                first = current

            result = target.send(current)
            current = yield result


@contextmanager
def lazy_counter(*counters, **kwargs):
    """Delay the update of the counter."""
    target = kwargs.get('target')

    local_counter = Counter()
    try:
        yield local_counter
    except BatchEndException:
        pass
    finally:
        if target:
            target.send(local_counter)

        for counter in counters:
            counter.update(local_counter)


@consumer
def delay(target, speedup=100.0, max_delay=2):
    last = None

    with closing(target):
        while True:
            tweet = yield

            if speedup and last is not None and last < tweet.created_at:
                delay = tweet.created_at - last
                delay = delay.total_seconds() / speedup
                delay = min([delay, max_delay])

                time.sleep(delay)

            last = tweet.created_at
            target.send(tweet.raw)


@consumer
def mutate(target, mutator=None):
    if mutator is None:
        def mutator(current, first):
            return current

    result = None

    with closing(target):
        first = current = yield result

        while True:
            result = target.send(mutator(current, first))
            current = yield result


@contextmanager
def closing(*targets, **kwargs):
    """Throw exceptions to targets and close targets on exit."""
    mutable_targets = kwargs.get('mutable_targets', [])
    try:
        yield
    finally:
        for target in chain(targets, mutable_targets):
            target.close()


class PropogatedException(Exception):
    """"""


class BatchEndException(PropogatedException):
    """
    Is thrown by batch() to the target generator on the end of the
    current batch.
    """

    def __init__(self, last_item, batch_size):
        self.last_item = last_item
        self.batch_size = batch_size


class SendNext(object):
    """
    Returned a consumer if the sent value is ignored, and the next value
    should be sent immediately.
    """



import codecs
import inspect
import logging
import sys

import opster

from poultry.config import Config
from poultry.producers import from_stream

logger = logging.getLogger(__name__)


class Dispatcher(opster.Dispatcher):
    def __init__(self, *globaloptions):
        globaloptions = (
            tuple(globaloptions) +
            (
                ('v', 'verbose', False, 'Be verbose.'),
                ('c', 'config', Config.default_config_file, 'Configuration file'),
                ('s', 'source', '', 'The tweet source.'),
                ('', 'utf8', False, 'Force UTF8 output.'),
            )
        )

        super(Dispatcher, self).__init__(
            globaloptions=globaloptions,
            middleware=_middleware,
        )


def _middleware(func):
    def wrapper(*args, **kwargs):

        if kwargs.pop('utf8'):
            # Always force utf8 output.
            sys.stdout = codecs.getwriter('utf8')(sys.stdout)

        if func.__name__ == 'help_inner':
            return func(*args, **kwargs)

        f_args = inspect.getargspec(func)[0]

        verbose = kwargs.pop('verbose')

        config = kwargs['config'] = Config(kwargs.get('config'))
        if 'config' not in f_args:
            kwargs.pop('config')

        source = kwargs.pop('source')
        producer = lambda target: from_stream(target, source, config)

        if 'producer' in f_args:
            kwargs['producer'] = producer

        logger = logging.getLogger('poultry')

        if verbose:
            handler = logging.StreamHandler()
            formatter = logging.Formatter('%(asctime)-6s: %(name)s - %(levelname)s - %(message)s')

            handler.setFormatter(formatter)
            logger.addHandler(handler)
            logger.setLevel(logging.DEBUG)
        else:
            logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.CRITICAL)

        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            pass
        except Exception:
            logger.error('Middleware captured an exception', exc_info=True)
            raise

    return wrapper



import json
import logging
import unicodedata
import sys

from collections import namedtuple
from datetime import datetime
from email.utils import parsedate_tz
from itertools import chain


logger = logging.getLogger(__name__)
TWEET_CREATED_AT_CHANGE_WARNING = False


class Tweet(object):
    def __init__(self, raw_json):
        try:
            tweet = json.loads(raw_json)
        except ValueError:
            raise TweetValueError("The passed json can't be parsed.")
        else:
            if isinstance(tweet, dict) and 'text' in tweet:
                self.raw = raw_json
                self.parsed = tweet
            else:
                raise TweetValueError("There is no 'text' field in the passed json.")

    @property
    def text(self):
        """The unprocessed text of the tweet."""
        return self.parsed['text']

    @property
    def hashtags(self):
        '''
        The lowercased hashtags that are in the tweet.
        '''
        return set(h['text'].lower() for h in self.parsed['entities']['hashtags'])

    @property
    def urls(self):
        '''
        The urls that are in the tweet.
        '''
        return set(u['url'] for u in self.parsed['entities']['urls'])

    @property
    def user_mentions(self):
        '''
        The mentioned users in the tweet.
        '''
        return set(m['screen_name'] for m in self.parsed['entities']['user_mentions'])

    @property
    def user_mention_ids(self):
        '''
        The IDs of the mentioned users in the tweet.
        '''
        return set(m['id'] for m in self.parsed['entities']['user_mentions'])

    @property
    def screen_name(self):
        try:
            return self.parsed['user']['screen_name']
        except KeyError:
            logger.warning("An error accessing ['user']['screen_name']")

    @property
    def user_id(self):
        return self.parsed['user']['id']

    @property
    def created_at(self):
        '''
        Tweet's creation time as a datetime object.
        '''
        # XXX a hack to make it possible to change the creation time
        # of a tweet.
        _created_at = getattr(self, '_created_at', None)
        if _created_at:
            return _created_at

        return self.orig_created_at

    @created_at.setter
    def created_at(self, value):
        global TWEET_CREATED_AT_CHANGE_WARNING

        if not TWEET_CREATED_AT_CHANGE_WARNING:
            logger.warn('{} created_at property is changed. '
                        'This change is not reflected in '
                        'self.raw!'
                        ''.format(type(self)))
            TWEET_CREATED_AT_CHANGE_WARNING = True

        self._created_at = value

    @property
    def orig_created_at(self):
        try:
            created_at = self.parsed['created_at']
        except KeyError:
            logger.warning("An error accessing ['created_at']")
        else:
            return self._created_at_to_datetime(created_at)

    @staticmethod
    def _created_at_to_datetime(created_at):
        '''
        Convert a date represented in the Twitter format to a datetime
        object.
        '''
        time_tuple = parsedate_tz(created_at)
        return datetime(*time_tuple[:6])

    @property
    def retweeted_status(self):
        try:
            return Tweet(json.dumps(self.parsed['retweeted_status']))
        except ValueError:
            pass
        except KeyError:
            pass

    @property
    def id(self):
        return self.parsed['id']

    @property
    def coordinates(self):
        """The location of the tweet.

        `None` is returned in case there is no geo information.

        """
        try:
            coor = self.parsed['coordinates']
        except KeyError:
            pass
        else:
            if coor and coor['type'] == 'Point':
                return Coordinates(*coor['coordinates'])

    @property
    def twitter_url(self):
        '''
        Tweet's url at twitter.com
        '''
        return 'https://twitter.com/#!/{t.screen_name}/status/{t.id}'.format(t=self)

    @property
    def text_without_entities(self):
        '''
        The text of the tweet without entities (hashtags, ursl and
        user mentions).
        '''
        entities = self.parsed['entities'].values()
        indicies = list(chain.from_iterable((e['indices'] for e in es) for es in entities))

        text = list(self.text)
        for start, end in indicies:
            length = end - start
            text[start:end] = [None] * length

        return u''.join(filter(None, text))

    @property
    def tokens(self):
        '''
        Tokenized text of the tweet.

        The minimal length of a token is 3. Only letters and numbers
        make up tokens.

        Hashtags, urls, and user mentions are not included.
        '''
        return self.get_tokens()

    def get_tokens(self, min_token_len=3, allowed_categories='LN'):
        """Tokenized text of the tweet.

        :param min_token_len: The minimal length of a token. Strings
                              less than this value are not considered
                              to be tokens.

        :param allowed_categories: The categories of which a string
                                   should contain in order to be
                                   considered as a token. Refer to the
                                   `unicodedata` module documentation.

        Hashtags, urls, and user mentions are not included.

        http://www.fileformat.info/info/unicode/category/index.htm

        """
        tokens = self.text_without_entities.split()
        tokens = (t.lower() for t in tokens)

        return filter(lambda s: len(s) >= min_token_len,
                      [u''.join(c for c in t if unicodedata.category(c)[0] in allowed_categories)
                       for t in tokens
                       ]
                      )

    @property
    def is_spam(self):
        '''
        Indication whether the tweet is considered as spam.

        .. todo: parameters instead of magic numbers.
        '''
        entity_violations = (2 < e for e in [self.hashtags, self.urls, self.user_mentions])
        lenght_violantion = len(self.tokens) < 5

        return any(entity_violations) or lenght_violantion

    def filter(self,
               follow=None,
               track=None,
               locations=None,
               start_date=None,
               end_date=None,
               ):
        '''
        Match the predicates to the tweet.

        :return: `True` if there is a match, `False` otherwise.

        Mimics Twitter `statuses/filter`_ method of the Streaming API.

        .. _statuses/filter: https://dev.twitter.com/docs/streaming-api/methods#statuses-filter
        '''

        if start_date is not None and self.created_at < start_date:
            return False

        if end_date is not None and end_date < self.created_at:
            return False

        follow = set(int(f) for f in follow) if follow is not None else set()

        try:
            unicode
        except NameError:
            unicode = str

        track = set(unicode(t.lower()) for t in track) if track is not None else set()
        locations = set(locations) if locations is not None else set()

        common_follow = follow.intersection(self.user_mentions) or self.user_id in follow
        common_track = (t.lower() in self.text.lower() for t in track)

        coor = self.coordinates
        place = self.parsed['place']
        if coor:
            common_location = (sw[0] <= coor.lon <= ne[0] and sw[1] <= coor.lat <= ne[1]
                               for sw, ne in locations)
        elif place and place['bounding_box']:
            # XXX Is it always a rectangle?
            bounding_box = place['bounding_box']['coordinates'][0]

            # XXX not the best way to intersect the given location and
            # the tweet's place polygon.
            sw_lon = min(lon for lon, lat in bounding_box)
            sw_lat = min(lat for lon, lat in bounding_box)
            ne_lon = max(lon for lon, lat in bounding_box)
            ne_lat = max(lat for lon, lat in bounding_box)

            common_location = (not (sw[0] > ne_lon or
                                    ne[0] < sw_lon or
                                    ne[1] > sw_lat or
                                    sw[1] < ne_lat)
                               for sw, ne in locations)

        else:
            common_location = []

        search_terms = ([follow, track, locations])
        common = any([common_follow,
                      any(common_track),
                      any(common_location),
                      ]
                     )

        return bool(not any(search_terms) or common)

    def __unicode__(self):
        return (
            u'{t.screen_name}: {t.text}\n'
            u'{t.twitter_url}\n'
            u'{geo}'
            u'{t.created_at}'
            ''.format(
                t=self,
                geo='http://www.openstreetmap.org/?mlat={c.lat}&mlon={c.lon}&zoom=6\n'.format(c=self.coordinates) if self.coordinates else '',
            )
        )

    def __repr__(self):
        return ('<{s.__class__.__name__}('
                'created_at={s.created_at}'
                ')>'.format(s=self)
                )

    if sys.version_info[0] >= 3: # Python 3
        def __str__(self):
            return self.__unicode__()
    else:  # Python 2
        def __str__(self):
            return self.__unicode__().encode('utf8')


class TweetValueError(ValueError):
    '''Thrown when a class:`Tweet` can't be built.'''


class Coordinates(namedtuple('Coordiantes', 'lon lat')):
    @staticmethod
    def from_string(s):
        def two_pairs(x):
            x = tuple(x)
            return x[:2], x[2:]

        return two_pairs(float(i.strip()) for i in s.split(','))



import logging
from time import sleep
from itertools import chain
from multiprocessing import Process

try:
    from multiprocessing import SimpleQueue
except ImportError:
    from multiprocessing.queues import SimpleQueue

try:
    from urllib.parse import quote
except ImportError:
    from urllib import quote

from sys import version_info
PY2 = version_info < (3, )


import requests
from requests_oauthlib import OAuth1Session

from poultry import consumers


logger = logging.getLogger(__name__)


class StreamProducer(Process):
    """A producer of a tweet stream retrieved from twitter.

    :param target: A coroutine to which collected tweets are sent.
    :param twitter_credentials: A dictionary with `access_token`,
    `access_token_secret`, `consumer_key` and `consumer_secret`.
    :param follow: A list of user ids to follow.
    :param track: A list of phrases to track.
    :param locations: A list of coordinates to get.

    ..todo:: Parameter description!

    The default access level allows up to 400 track keywords, 5,000
    follow userids and 25 0.1-360 degree location boxes.
    https://dev.twitter.com/docs/streaming-api/methods

    """

    def __init__(self, target, twitter_credentials,
                 follow=None, track=None, locations=None,
                 url='https://stream.twitter.com/1.1/statuses/filter.json',
                 *args, **kwargs):
        super(StreamProducer, self).__init__(*args, **kwargs)

        self.target = target

        self.track = track if track is not None else []
        self.follow = follow if follow is not None else []
        self.locations = locations if locations is not None else []
        self.url = url

        self.client = OAuth1Session(
            twitter_credentials['consumer_key'],
            client_secret=twitter_credentials['consumer_secret'],
            resource_owner_key=twitter_credentials['access_token'],
            resource_owner_secret=twitter_credentials['access_token_secret'],
        )

    def _run(self):
        target = self.target

        def _quote(items):
            # XXX It's not clear for me how the parameters have to be quote.
            if PY2:
                items = (i.encode('utf-8') for i in items)
            return quote(','.join(items), safe=', ')

        data = {
            p: _quote(getattr(self, p)) for p in 'track follow'.split() if getattr(self, p)
        }

        locations = ','.join(str(f) for f in chain.from_iterable(chain.from_iterable(self.locations)))
        if locations:
            data['locations'] = locations
        response = self.client.post(self.url, data=data, stream=True)

        response.raise_for_status()

        line = None
        for line in response.iter_lines():
            target.send(line.decode('utf-8'))
        else:
            # XXX Should be changed to something meaningful
            raise EndOfStreamError(line)

    def run(self):
        try:
            while True:
                try:
                    self._run()
                # XXX implement meaningful reconnection strategy.
                #     https://dev.twitter.com/docs/streaming-apis/connecting#Reconnecting
                except requests.HTTPError:
                    logger.warn('An http error occurred. Reconnecting...', exc_info=True)
                    sleep(10)
                except EndOfStreamError:
                    logger.warn('The stream ended. Reconnecting...', exc_info=True)
                    sleep(60)
        except KeyboardInterrupt:
            pass


class StreamConsumer(Process):
    """A consumer of a stream of tweets."""
    def __init__(self, queue, target, *args, **kwargs):
        super(StreamConsumer, self).__init__(*args, **kwargs)

        self.queue = queue
        self.target = target

    def run(self):
        try:
            from_simple_queue(self.target, self.queue)
        except KeyboardInterrupt:
            pass


def from_simple_queue(target, queue):
    with consumers.closing(target):
        while True:
            item = queue.get()

            if item is StopIteration:
                break

            target.send(item)


def from_twitter_api(target, endpoint, config):
    """Consume tweets from a Streaming API endpoint."""
    endpoint_to_url = {
        'twitter://sample': 'https://stream.twitter.com/1.1/statuses/sample.json',
        'twitter://filter': 'https://stream.twitter.com/1.1/statuses/filter.json',
    }

    if endpoint == 'twitter://filter':
        filter_predicates = config.global_filter.predicates
        kwargs = {
            'follow': filter_predicates['follow'],
            'track': filter_predicates['track'],
            'locations': filter_predicates['locations'],
        }
    else:
        kwargs = {}

    # The communication point of the consumer and producer processes.
    queue = SimpleQueue()

    # Create the producer first to be sure that it exists before creating and
    # starting the consumer.
    producer = StreamProducer(
        twitter_credentials=dict(config.items('twitter')),
        target=consumers.to_simple_queue(queue),
        url=endpoint_to_url[endpoint],
        **kwargs
    )

    # start the consumer
    consumer = StreamConsumer(queue, target)
    consumer.start()

    try:
        producer.start()
        producer.join()
    finally:
        # Tell the consumer to stop
        queue.put(StopIteration)
        consumer.join()


class EndOfStreamError(IOError):
    """Twitter streams are supposed to be infinite.

    The exception should be thrown, if for any reason the stream has ended.

    """



import os
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


dirname = os.path.dirname(__file__)


with open(os.path.join(dirname, 'README.rst')) as f:
    long_description = f.read()
with open(os.path.join(dirname, 'CHANGES.rst')) as f:
    long_description += '\n' + f.read()


class Tox(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['--recreate']
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import tox
        errno = tox.cmdline(self.test_args)
        sys.exit(errno)


setup(
    name='poultry',
    version='1.1.2',
    description='A tweet collection manager.',
    long_description=long_description,
    # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Topic :: Utilities',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
    keywords='',
    author='Dmitrijs Milajevs',
    author_email='dimazest@gmail.com',
    url='https://github.com/dimazest/poultry',
    license='MIT license',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'opster',
        'setuptools',
        'requests',
        'requests-oauthlib',
    ],
    entry_points={
        'console_scripts': [
            'poultry = poultry.main:dispatch',
        ],
    },
    tests_require=['tox'],
    cmdclass={'test': Tox},
)



import heapq
import logging

from poultry.consumers import closing, consumer


logger = logging.getLogger(__name__)


def merge(pipes, provider=None):
    m = _merge(pipes, provider)

    while True:
        try:
            next(m)
        except StopIteration:
            break


@consumer
def _merge(pipes, provider=None):
    epipes = list(enumerate(pipes))

    merger = _merger(
        {tag: t for tag, (_, t) in epipes},
        provider=provider,
    )

    tagged_sources = {tag: s(_tagger(merger, tag)) for tag, (s, _) in epipes}

    conductor = _conductor(merger, tagged_sources)

    while True:
        r = next(conductor)
        if r is not None:
            item, _ = r
            yield item


@consumer
def _conductor(merger, tagged_sources):
    result = None

    with closing(merger):
        # The first step: push each target once.
        for source in tagged_sources.values():
            try:
                result = next(source)
                if result is None:
                    yield
            except StopIteration:
                pass

        item, source_priority = result

        # The second step: push targets as merge requests.
        while True:
            yield item, source_priority

            while source_priority:
                tag = source_priority.pop(0)
                try:
                    item, source_priority = next(tagged_sources[tag])
                except StopIteration:
                    pass
                else:
                    break
            else:
                return


@consumer
def _merger(tagged_targets, provider=None):
    if provider is None:
        def provider(current):
            return current

    targets = set(tagged_targets.values())
    with closing(*targets):

        items = []
        heapq.heapify(items)

        applied_sources = set()
        closed_sources = set()
        all_sources = set(tagged_targets)

        try:
            # First step: get at least one item from each source
            while applied_sources != all_sources:
                try:
                    tag, item = yield
                except TaggedGeneratorExit as e:
                    applied_sources.add(e.tag)
                    closed_sources.add(e.tag)
                else:
                    heapq.heappush(items, (provider(item), tag, item))
                    applied_sources.add(tag)

            # Second step: push forward the smallest item to its target
            # and push back to conductor the priorities of the sources.
            while closed_sources != applied_sources:
                priorities = [i[1] for i in sorted(items)]

                _, tag, item = heapq.heappop(items)

                result = tagged_targets[tag].send(item)

                try:
                    tag, item = yield result, priorities
                except TaggedGeneratorExit as e:
                    closed_sources.add(e.tag)
                    logger.debug('%s is closed', tag)
                else:
                    heapq.heappush(items, (provider(item), tag, item))

        finally:
            for _, tag, item in sorted(items):
                tagged_targets[tag].send(item)


@consumer
def _tagger(target, tag):
    """Send to the target pairs (tag, item)."""
    # Target is not being closed!
    result = None

    while True:
        try:
            item = yield result
        except GeneratorExit as e:
            target.throw(TaggedGeneratorExit(tag))
            raise e
        else:
            result = target.send((tag, item))


class TaggedGeneratorExit(Exception):
    def __init__(self, tag, *args, **kwargs):
        super(TaggedGeneratorExit, self).__init__(*args, **kwargs)

        self.tag = tag



import os
import platform
from itertools import chain


if platform.python_implementation == 'CPython':
    import ujson as json
else:
    import json


def get_file_names(input_dir):
    if input_dir:
        file_names = sorted(chain.from_iterable((os.path.join(p, f) for f in fs) for p, _, fs in os.walk(input_dir)))
    else:
        file_names = []

    for f in file_names:
        yield f
